Require Export ZArith Streams List String Bool.
Open Scope Z_scope.

Extract Inductive list  =>  "list" [ "[]" "(::)" ].

Extract Inductive option => "option" [ "Some" "None" ].

Extract Inductive unit => "unit" [ "()" ].

Extract Inductive bool => "bool" ["true" "false"].

Definition pattern_match_number (A :Type) 
  (case0 : unit -> A) (case1 : unit -> A)
  (caseneg : Z -> A) (caseodd : Z -> A) (caseeven : Z -> A) (z:Z) : A:=
  match z with
    0%Z => case0 tt
  | 1%Z => case1 tt
  | Zpos (xI p) => caseodd (Zpos p)
  | Zpos (xO p) => caseeven (Zpos p)
  | Zneg p => caseneg (Zpos p)
  end.

Definition div10_cont (A:Type)(cont: Z -> Z -> A) (n:Z) : A :=
  let (q,r) := Zdiv_eucl n 10 in cont q r.

Parameter n_string : Type.
Extract Constant n_string => "String.t".

Parameter n_array : Type -> Type .
Extract Constant n_array "'a" => "'a array".

Parameter sys_argv : n_array n_string.
Extract Constant sys_argv => "Sys.argv".

Axiom array_to_list : forall (A:Type)(a : n_array A), list A.
Extract Constant array_to_list => "Array.to_list".

Import Ascii.

Definition decode_char (A:Type)(f : bool -> A -> A) (a:A) (c:ascii) : A :=
 match c with
   Ascii b1 b2 b3 b4 b5 b6 b7 b8 =>
   f b1 (f b2 (f b3 (f b4 (f b5 (f b6 (f b7 (f b8 a)))))))
 end.

Definition ascii_of_z (x:Z):ascii :=
  match x with
    0 => ascii_of_nat 0
  | Zpos p => ascii_of_pos p
  | Zneg p => ascii_of_pos p
  end.

Axiom native_to_string_aux : forall (zero : Z)(one : Z)
  (twicep1 twice : Z -> Z) (string_cons : Z -> string -> string)
  (empty_string : string) (s : n_string), string.
Extract Constant native_to_string_aux => "Conversion.native_to_string".
Definition native_to_string :=
  native_to_string_aux 0 1 (fun x=>2*x+1) (Zmult 2) 
    (fun x => String (ascii_of_z x)) EmptyString.

Definition argument_list :=
  List.map native_to_string (array_to_list _ sys_argv).

Axiom z_of_string_aux : forall (zero : Z)(one :Z)(twicep1 twice : Z -> Z)
 (zmult zplus : Z -> Z -> Z) (zopp : Z -> Z) (s : n_string), Z.
Extract Constant z_of_string_aux =>  "Conversion.z_of_string".
Definition z_of_n_string := 
  z_of_string_aux 0 1 (fun x => 2*x+1) (Zmult 2) Zmult Zplus Zopp.

Axiom read_line : unit -> n_string.
Extract Constant read_line => "read_line".

CoFixpoint input_stream_Z (v:unit): Stream Z :=
  Cons (z_of_n_string (read_line tt)) (input_stream_Z v).

Axiom output_stream_Z :  (Stream Z) -> unit.
Extract Constant output_stream_Z =>
  "let rec f s = let Cons(a, s') =
    try Lazy.force s with End_of_file -> exit 0 in
    (Conversion.print_z pattern_match_number pattern_match_number
        div10_cont a; print_string ""\n""; f s')  in f".

Axiom n_int : Type.
Extract Constant n_int => "int".
Parameter int0 : n_int.
Extract Constant int0 => "0".
Parameter int_twice_carry : bool -> n_int -> n_int.
Extract Constant int_twice_carry =>
  "fun b x -> 2 * x + (if b then 1 else 0)".

Definition ascii_to_int (c:ascii) : n_int :=
 Eval compute in
 let f := int_twice_carry in
 match c with
   Ascii b1 b2 b3 b4 b5 b6 b7 b8 =>
    f b1 (f b2 (f b3 (f b4 (f b5 (f b6 (f b7 (f b8 int0)))))))
  end.

Axiom output_char : n_int -> unit.
Extract Constant output_char => "Conversion.output_char".
Axiom flush_stdout : unit -> unit.
Extract Constant flush_stdout => "Conversion.flush_stdout".

Fixpoint output_string (s : string) :=
  match s with
    EmptyString => flush_stdout tt
  | String c tl =>
    let () := output_char (ascii_to_int c) in output_string tl
  end.

Axiom output_stream_string : forall (s:Stream string), unit.
Extract Constant output_stream_string => "
  let rec f s =
    match Lazy.force s with
      Cons(s1,tl) -> (output_string s1; f tl) in
    f".

Open Scope string_scope.


Definition is_digit (a : ascii) : bool :=
  match a with
    Ascii b1 b2 b3 b4 b5 b6 b7 b8 => 
       andb (negb b8) (andb (negb b7) 
           (andb b6 (andb b5
              (orb (negb b4) (andb b4 (andb (negb b3) (negb b2)))))))
  end.

Definition digit_value (a : ascii) : Z :=
   Eval compute in 
    let f := fun (b:bool) (x:Z) => 2*x + if b then 1 else 0 in
    match a with
      Ascii b1 b2 b3 b4 _ _ _ _ =>
        f b1 (f b2 (f b3 (f b4 0)))
    end.

Fixpoint z_of_string1 (seed : Z) (s : string) {struct s} : Z :=
  match s with
    EmptyString => seed
  | String a s' => 
    if is_digit a then z_of_string1 (10 * seed + digit_value a) s'
    else seed
  end.

Definition z_of_string (s : string) : Z := z_of_string1 0 s.

Definition digit_to_char (x : Z) : ascii :=
   match x with 
     Z0 => ascii_of_pos 48
   | Zpos p => ascii_of_pos (48 + p)
   | Zneg p => ascii_of_pos 48
   end.

Fixpoint string_of_z1 (x : Z) (p:positive) {struct p} : list ascii :=
  match p, x with
    _, Z0 => nil
  | xH, x => digit_to_char x::nil
  | xI p, _ => let (q,r) := Zdiv_eucl x 10 in digit_to_char r::string_of_z1 q p
  | xO p, _ => let (q,r) := Zdiv_eucl x 10 in digit_to_char r::string_of_z1 q p
  end.

Fixpoint string_of_rev_list (l:list ascii) (acc: string) {struct l} : string :=
  match l with
    nil => acc | a::tl => string_of_rev_list tl (String a acc)
  end.

Definition string_of_z (x:Z) :=
  match x with
    Z0 => "0"
  | Zpos p => string_of_rev_list (string_of_z1 x p) EmptyString
  | Zneg p => "-"++string_of_rev_list (string_of_z1 (Zpos p) p) EmptyString
  end.

Axiom input_to_n_string : unit -> n_string.
Extract Constant input_to_n_string => "Conversion.input_to_string".

Definition input_to_string (v : unit) : string :=
  native_to_string (input_to_n_string v).

CoFixpoint input_stream_string (v:unit) : Stream string :=
  Cons (native_to_string (read_line tt)) (input_stream_string tt).