Require Import ZArith String List extracted_io.

Inductive aexpr : Type :=
  avar (s : string) | anum (n :Z) | aplus (a1 a2:aexpr) | amult (a1 a2:aexpr).

Bind Scope aexpr_scope with aexpr.
Notation "X + Y" := (aplus X Y) : aexpr_scope.
Notation "X * Y" := (amult X Y) : aexpr_scope.
Coercion anum : Z >-> aexpr.

Inductive bexpr : Type :=  blt (e1 e2 : aexpr).
Bind Scope bexpr_scope with bexpr.
Notation "X < Y" := (blt X Y) : bexpr_scope.
Delimit Scope bexpr_scope with B.

Inductive instr : Type :=
  assign (x : string) (e : aexpr) | sequence (i1 i2: instr)
 | while (b :bexpr)(i : instr).

Inductive assert : Type :=
  a_b(b: bexpr) | a_not(a: assert) | a_conj(a a': assert)
| pred(s: string)(l: list aexpr) | a_true | a_false.

Notation "X /\ Y" := (a_conj X Y) : assert_scope.
Notation "~ X" := (a_not X) : assert_scope.
Delimit Scope assert_scope with A.

Inductive condition : Type := c_imp (a1 a2 : assert).

(* output functions on these data types. *)

Fixpoint aexpr_to_string (e : aexpr) : string :=
  match e with
    avar x => x
  | anum n => string_of_z n
  | aplus e1 ((aplus _ _) as e2) => 
    aexpr_to_string e1 ++ " + (" ++ aexpr_to_string e2 ++ ")"
  | aplus e1 e2 => aexpr_to_string e1 ++ " + " ++ aexpr_to_string e2
  | amult e1 e2 => aexpr_to_string e1 ++ " * " ++ aexpr_to_string e2
  end.

Fixpoint env_to_string (r : list(string*Z)) : string :=
  match r with
    nil => "nil"
  | (x, n)::tl => "(" ++ x ++ "," ++ string_of_z n ++ ")." ++ env_to_string tl
  end.

(* Can you write a similar function for instructions? *)

Bind Scope assert_scope with assert.

Coercion a_b : bexpr >-> assert.

Open Scope string_scope.
Coercion avar : string >-> aexpr.

Inductive a_instr : Type :=
  a_assign (x:string) (e:aexpr) | a_seq (i1 i2: a_instr)
| a_while (b:bexpr) (inv:assert) (i:a_instr)
| prec (a:assert)(i:a_instr).

Fixpoint un_annot (i:a_instr) : instr :=
  match i with
    a_assign x e => assign x e
  | a_seq i1 i2 => sequence (un_annot i1) (un_annot i2)
  | a_while b a i => while b (un_annot i)
  | prec a i => un_annot i
  end.