Require Import ZArith List String syntax parser axiom.

(* Les deux lemmes suivants sont importants pour la dmonstration  faire. *)

Check valid_app.
Check valid_triv.

Lemma vcg_correct :
  forall m i Q, valid_l m (vc i Q) -> ax_sem m (pc i Q) (un_annot i) Q.
Proof.
 intros m i.
 induction i.
(* A FAIRE: le reste de cette dmonstration. *)
Admitted.

(* Voici maintenant un exemple d'utilisation.
  Le programme pg1 calcule la somme des entiers de 1  10000,
  le rsultat est prvisible par une preuve mathmatique.

  La clause "Eval vm_compute in" est utilise pour s'assurer que 
  l'analyse syntaxique est effectue au moment de la dfinition, et qu'elle
  ne sera plus ncessaire. *)

Definition pg1 :=
  Eval vm_compute in
  parse_ai' 
    "[eq(x, 0) /\ eq (y, 0) ]
     while x < 10000 do
       [leq(x, 10000) /\ eq(2*y, x*(x+1))]
       x:=x+1; y:=y+x
     done".

(* Vrifions qu'il n'y a pas d'erreur  l'analyse syntaxique. *)

Eval unfold pg1 in pg1.

(* Dans les annotations nous utilisons deux prdicats "leq(..)" et "eq(..)",
  Le sens a donner  ces prdicats est prcis par les deux dfinitions
  suivantes. *)
  

Definition p_le := fun l =>
  match l with x::y::nil => (x <= y)%Z | _ => False end.

Definition p_eq := fun (l:list Z) =>
  match l with x::y::nil => x = y | _ => False end.

(* Pour l'interprtation des assertions, nous utiliserons donc le
  vocabulaire suivant. *)

Definition m := ("leq", p_le)::("eq", p_eq)::nil.

(* Voici quelques notations pour amliorer la lisibilit des assertions. *)

Notation "x <= y" :=
  (pred "leq" (@cons aexpr x (@cons aexpr y nil))) : assert_scope.
Notation "x = y" :=
  (pred "eq" (@cons aexpr x (@cons aexpr y nil))) : assert_scope.
Notation "x < y" := (a_b (blt x y)) : assert_scope.
Delimit Scope aexpr_scope with E.

(* Cette tactique est  utiliser sur tous les buts de la forme
  "valid_l m ...", pour les transformer en de simples buts arithmtique
  de Coq. *)
Ltac expand_semantics :=
 lazy beta zeta iota delta [valid_l fold_right vc pc app a_subst subst b_subst
  l_subst valid m i_a i_c f_p string_dec string_rec
  string_rect sumbool_rec sumbool_rect Ascii.ascii_dec Ascii.ascii_rec
  Ascii.ascii_rect Bool.bool_dec bool_rec bool_rect eq_rec_r
  eq_rec eq_rect sym_eq p_eq p_le lf' af' bf'].

(* Voici un exemple de preuve. *)
Lemma pg1_res :
  ax_sem m ("x" = 0 /\ "y" = 0)%A (un_annot pg1) ("y" = 50005000).
 apply (vcg_correct m pg1 ("y" = 50005000)).
 unfold pg1; expand_semantics.
(* A partir de l, on n'a plus qu' prouver des proprits arithmtiques. *)
 split.
  intros g ; generalize (g "x") (g "y"); intros x y [X Y].
  rewrite X; rewrite Y. omega.
 split.
  intros g; generalize (g "x") (g "y"); clear g; intros x y.
  intros H; assert (Hx : x = 10000) by omega.
  rewrite Hx in H; apply Zmult_reg_l with 2.
   discriminate.
  destruct H as [H1 [H2 H3]]; rewrite H3; auto.
 split; auto.
 intros g; generalize (g "x") (g "y"); clear g; intros x y.
 intros [xlt [_ Y]].
 split.
  omega.
 rewrite (Zmult_plus_distr_r 2).
 rewrite Y.
 ring.
Qed.


(* A FAIRE:
  Trouvez les annotations  ajouter au programme pg2 pour montrer qu'il calcule
  bien le quotient et le reste de la division de x par y. *)

Definition pg2 :=
  Eval vm_compute in 
  parse_i
    "r := x; q := 0;
     while y < r + 1 do
       r := -1 * y + r;
       q := 1 + q
     done".

Definition pg2' :=
  Eval vm_compute in
  parse_ai' "dummy".

Eval vm_compute in pg2'.

Definition pg2_final_spec : assert :=
   (("x" = ("q" * "y" + "r")%E)%A /\ 0 <= "r" /\ ("r" < "y"))%A.

Lemma pg2_res : ax_sem m (pc pg2' pg2_final_spec) pg2 pg2_final_spec.
(* L'tape suivante devrait passer sans problme si vous n'avez pas cass
 le programme en ajoutant les annotations. *)
change pg2 with (un_annot pg2').
apply vcg_correct.
unfold pg2, pg2', pg2_final_spec; expand_semantics.
(* A FAIRE : le reste de cette dmonstration. *)
Qed.
